@@ -7,6 +7,8 @@ import asyncio
 import json
 import xyzservices
 from datetime import date, timedelta
+from math import isnan
+import warnings
 
 from ipywidgets import (
     Widget, DOMWidget, Box, Color, CallbackDispatcher, widget_serialization,
@@ -59,7 +61,7 @@ def basemap_to_tiles(basemap, day=yesterday, **kwargs):
 
     return TileLayer(
         url=url,
-        max_zoom=basemap.get('max_zoom', 19),
+        max_zoom=basemap.get('max_zoom', 18),
         min_zoom=basemap.get('min_zoom', 1),
         attribution=basemap.get('html_attribution', '') or basemap.get('attribution', ''),
         name=basemap.get('name', ''),
@@ -78,6 +80,11 @@ def wait_for_change(widget, value):
     return future
 
 
+class PaneException(TraitError):
+    """Custom PaneException class."""
+    pass
+
+
 class LayerException(TraitError):
     """Custom LayerException class."""
     pass
@@ -110,6 +117,8 @@ class Layer(Widget, InteractMixin):
         Custom name for the layer, which will be used by the LayersControl.
     popup: object
         Interactive widget that will be shown in a Popup when clicking on the layer.
+    pane: string
+        Name of the pane to use for the layer.
     """
 
     _view_name = Unicode('LeafletLayerView').tag(sync=True)
@@ -127,6 +136,7 @@ class Layer(Widget, InteractMixin):
     popup_min_width = Int(50).tag(sync=True)
     popup_max_width = Int(300).tag(sync=True)
     popup_max_height = Int(default_value=None, allow_none=True).tag(sync=True)
+    pane = Unicode('').tag(sync=True)
 
     options = List(trait=Unicode()).tag(sync=True)
 
@@ -549,16 +559,18 @@ class TileLayer(RasterLayer):
     url: string, default "https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png"
         Url to the tiles service.
     min_zoom: int, default 0
-        Minimum zoom for this tile service.
+        The minimum zoom level down to which this layer will be displayed (inclusive).
     max_zoom: int, default 18
-        Maximum zoom for this tile service.
-    min_native_zoom: int, default 0
-    max_native_zoom: int, default 18
+        The maximum zoom level up to which this layer will be displayed (inclusive).
+    min_native_zoom: int, default None
+        Minimum zoom number the tile source has available. If it is specified, the tiles on all zoom levels lower than min_native_zoom will be loaded from min_native_zoom level and auto-scaled.
+    max_native_zoom: int, default None
+        Maximum zoom number the tile source has available. If it is specified, the tiles on all zoom levels higher than max_native_zoom will be loaded from max_native_zoom level and auto-scaled.
     bounds: list or None, default None
         List of SW and NE location tuples. e.g. [(50, 75), (75, 120)].
     tile_size: int, default 256
         Tile sizes for this tile service.
-    attribution: string, default "Map data (c) <a href="https://openstreetmap.org">OpenStreetMap</a> contributors"
+    attribution: string, default None.
         Tiles service attribution.
     no_wrap: boolean, default False
         Whether the layer is wrapped around the antimeridian.
@@ -580,12 +592,11 @@ class TileLayer(RasterLayer):
     url = Unicode('https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png').tag(sync=True)
     min_zoom = Int(0).tag(sync=True, o=True)
     max_zoom = Int(18).tag(sync=True, o=True)
-    min_native_zoom = Int(0).tag(sync=True, o=True)
-    max_native_zoom = Int(18).tag(sync=True, o=True)
+    min_native_zoom = Int(default_value=None, allow_none=True).tag(sync=True, o=True)
+    max_native_zoom = Int(default_value=None, allow_none=True).tag(sync=True, o=True)
     bounds = List(default_value=None, allow_none=True, help='list of SW and NE location tuples').tag(sync=True, o=True)
     tile_size = Int(256).tag(sync=True, o=True)
-    attribution = Unicode('Map data (c) <a href="https://openstreetmap.org">OpenStreetMap</a> contributors').tag(
-        sync=True, o=True)
+    attribution = Unicode(default_value=None, allow_none=True).tag(sync=True, o=True)
     detect_retina = Bool(False).tag(sync=True, o=True)
     no_wrap = Bool(False).tag(sync=True, o=True)
     tms = Bool(False).tag(sync=True, o=True)
@@ -1088,18 +1099,78 @@ class LayerGroup(Layer):
     def add_layer(self, layer):
         """Add a new layer to the group.
 
+        .. deprecated :: 0.17.0
+           Use add method instead.
+
         Parameters
         ----------
         layer: layer instance
             The new layer to include in the group.
         """
+        warnings.warn("add_layer is deprecated, use add instead", DeprecationWarning)
+
+        self.add(layer)
+
+    def remove_layer(self, rm_layer):
+        """Remove a layer from the group.
+
+        .. deprecated :: 0.17.0
+           Use remove method instead.
+
+        Parameters
+        ----------
+        layer: layer instance
+            The layer to remove from the group.
+        """
+        warnings.warn("remove_layer is deprecated, use remove instead", DeprecationWarning)
+
+        self.remove(rm_layer)
+
+    def substitute_layer(self, old, new):
+        """Substitute a layer with another one in the group.
+
+        .. deprecated :: 0.17.0
+           Use substitute method instead.
+
+        Parameters
+        ----------
+        old: layer instance
+            The layer to remove from the group.
+        new: layer instance
+            The new layer to include in the group.
+        """
+        warnings.warn("substitute_layer is deprecated, use substitute instead", DeprecationWarning)
+
+        self.substitute(old, new)
+
+    def clear_layers(self):
+        """Remove all layers from the group.
+
+        .. deprecated :: 0.17.0
+           Use clear method instead.
+
+        """
+
+        warnings.warn("clear_layers is deprecated, use clear instead", DeprecationWarning)
+
+        self.layers = ()
+
+    def add(self, layer):
+        """Add a new layer to the group.
+
+        Parameters
+        ----------
+        layer: layer instance
+            The new layer to include in the group.
+        """
+
         if isinstance(layer, dict):
             layer = basemap_to_tiles(layer)
         if layer.model_id in self._layer_ids:
             raise LayerException('layer already in layergroup: %r' % layer)
         self.layers = tuple([layer for layer in self.layers] + [layer])
 
-    def remove_layer(self, rm_layer):
+    def remove(self, rm_layer):
         """Remove a layer from the group.
 
         Parameters
@@ -1107,11 +1178,12 @@ class LayerGroup(Layer):
         layer: layer instance
             The layer to remove from the group.
         """
+
         if rm_layer.model_id not in self._layer_ids:
             raise LayerException('layer not on in layergroup: %r' % rm_layer)
         self.layers = tuple([layer for layer in self.layers if layer.model_id != rm_layer.model_id])
 
-    def substitute_layer(self, old, new):
+    def substitute(self, old, new):
         """Substitute a layer with another one in the group.
 
         Parameters
@@ -1127,7 +1199,7 @@ class LayerGroup(Layer):
             raise LayerException('Could not substitute layer: layer not in layergroup.')
         self.layers = tuple([new if layer.model_id == old.model_id else layer for layer in self.layers])
 
-    def clear_layers(self):
+    def clear(self):
         """Remove all layers from the group."""
         self.layers = ()
 
@@ -1164,6 +1236,7 @@ class GeoJSON(FeatureGroup):
 
     data = Dict().tag(sync=True)
     style = Dict().tag(sync=True)
+    visible = Bool(True).tag(sync=True)
     hover_style = Dict().tag(sync=True)
     point_style = Dict().tag(sync=True)
     style_callback = Any()
@@ -1330,6 +1403,12 @@ class Choropleth(GeoJSON):
         The colormap used for the effect.
     key_on: string, default "id"
         The feature key to use for the colormap effect.
+    nan_color: string, default "black"
+        The color used for filling polygons with NaN-values data.
+    nan_opacity : float, default 0.4
+        The opacity used for NaN data polygons, between 0. (fully transparent) and 1. (fully opaque).
+    default_opacity: float, default 1.0
+        The opacity used for well-defined data (non-NaN values), between 0. (fully transparent) and 1. (fully opaque).
     """
 
     geo_data = Dict()
@@ -1338,8 +1417,11 @@ class Choropleth(GeoJSON):
     value_max = CFloat(None, allow_none=True)
     colormap = Any()
     key_on = Unicode('id')
+    nan_color = Unicode('black')
+    nan_opacity = CFloat(0.4)
+    default_opacity = CFloat(1.0)
 
-    @observe('style', 'style_callback', 'value_min', 'value_max', 'geo_data', 'choro_data', 'colormap')
+    @observe('style', 'style_callback', 'value_min', 'value_max', 'nan_color', 'nan_opacity', 'default_opacity', 'geo_data', 'choro_data', 'colormap')
     def _update_data(self, change):
         self.data = self._get_data()
 
@@ -1355,7 +1437,8 @@ class Choropleth(GeoJSON):
     def _default_style_callback(self):
         def compute_style(feature, colormap, choro_data):
             return dict(
-                fillColor=colormap(choro_data),
+                fillColor=self.nan_color if isnan(choro_data) else colormap(choro_data),
+                fillOpacity=self.nan_opacity if isnan(choro_data) else self.default_opacity,
                 color='black',
                 weight=0.9
             )
@@ -1366,12 +1449,15 @@ class Choropleth(GeoJSON):
         if not self.geo_data:
             return {}
 
+        choro_data_values_list = [x for x in self.choro_data.values() if not isnan(x)]
+
         if self.value_min is None:
-            self.value_min = min(self.choro_data.items(), key=lambda x: x[1])[1]
+            self.value_min = min(choro_data_values_list)
         if self.value_max is None:
-            self.value_max = max(self.choro_data.items(), key=lambda x: x[1])[1]
+            self.value_max = max(choro_data_values_list)
 
         colormap = self.colormap.scale(self.value_min, self.value_max)
+
         data = copy.deepcopy(self.geo_data)
 
         for feature in data['features']:
@@ -1829,6 +1915,10 @@ class LegendControl(Control):
 
     A control which contains a legend.
 
+    .. deprecated :: 0.17.0
+       The constructor argument 'name' is deprecated, use the 'title' argument instead.
+
+
     Attributes
     ----------
     title: str, default 'Legend'
@@ -1846,41 +1936,74 @@ class LegendControl(Control):
         "value 2": "#55A",
         "value 3": "#005"}).tag(sync=True)
 
-    def __init__(self, legend, *args, name="Legend", **kwargs):
+    def __init__(self, legend, *args, **kwargs):
+        kwargs["legend"] = legend
+        # For backwards compatibility with ipyleaflet<=0.16.0
+        if 'name' in kwargs:
+            warnings.warn("the name argument is deprecated, use title instead", DeprecationWarning)
+            kwargs.setdefault('title', kwargs['name'])
+            del kwargs['name']
         super().__init__(*args, **kwargs)
-        self.title = name
-        self.legend = legend
 
     @property
     def name(self):
+        """The title of the legend.
+
+        .. deprecated :: 0.17.0
+           Use title attribute instead.
+        """
+        warnings.warn(".name is deprecated, use .title instead", DeprecationWarning)
         return self.title
 
     @name.setter
     def name(self, title):
+        warnings.warn(".name is deprecated, use .title instead", DeprecationWarning)
         self.title = title
 
     @property
     def legends(self):
+        """The legend information.
+
+        .. deprecated :: 0.17.0
+           Use legend attribute instead.
+        """
+
+        warnings.warn(".legends is deprecated, use .legend instead", DeprecationWarning)
         return self.legend
 
     @legends.setter
     def legends(self, legends):
+        warnings.warn(".legends is deprecated, use .legend instead", DeprecationWarning)
         self.legend = legends
 
     @property
     def positioning(self):
+        """The position information.
+
+        .. deprecated :: 0.17.0
+           Use position attribute instead.
+        """
+        warnings.warn(".positioning is deprecated, use .position instead", DeprecationWarning)
         return self.position
 
     @positioning.setter
     def positioning(self, position):
+        warnings.warn(".positioning is deprecated, use .position instead", DeprecationWarning)
         self.position = position
 
     @property
     def positionning(self):
+        """The position information.
+
+        .. deprecated :: 0.17.0
+           Use position attribute instead.
+        """
+        warnings.warn(".positionning is deprecated, use .position instead", DeprecationWarning)
         return self.position
 
     @positionning.setter
     def positionning(self, position):
+        warnings.warn(".positionning is deprecated, use .position instead", DeprecationWarning)
         self.position = position
 
     def add_legend_element(self, key, value):
@@ -2011,9 +2134,9 @@ class Map(DOMWidget, InteractMixin):
         The current center of the map.
     zoom: float, default 12
         The current zoom value of the map.
-    max_zoom: int, default 18
+    max_zoom: float, default None
         Maximal zoom value.
-    min_zoom: int, default 1
+    min_zoom: float, default None
         Minimal zoom value.
     zoom_snap: float, default 1
         Forces the map’s zoom level to always be a multiple of this.
@@ -2071,10 +2194,9 @@ class Map(DOMWidget, InteractMixin):
 
     # Map options
     center = List(def_loc).tag(sync=True, o=True)
-    zoom_start = CFloat(12).tag(sync=True, o=True)
-    zoom = CFloat(12).tag(sync=True, o=True)
-    max_zoom = CFloat(18).tag(sync=True, o=True)
-    min_zoom = CFloat(1).tag(sync=True, o=True)
+    zoom = CFloat(default_value=None, allow_none=True).tag(sync=True, o=True)
+    max_zoom = CFloat(default_value=None, allow_none=True).tag(sync=True, o=True)
+    min_zoom = CFloat(default_value=None, allow_none=True).tag(sync=True, o=True)
     zoom_delta = CFloat(1).tag(sync=True, o=True)
     zoom_snap = CFloat(1).tag(sync=True, o=True)
     interpolation = Unicode('bilinear').tag(sync=True, o=True)
@@ -2137,6 +2259,7 @@ class Map(DOMWidget, InteractMixin):
     right = Float(0, read_only=True).tag(sync=True)
     left = Float(9007199254740991, read_only=True).tag(sync=True)
 
+    panes = Dict().tag(sync=True)
     layers = Tuple().tag(trait=Instance(Layer), sync=True, **widget_serialization)
 
     @default('layers')
@@ -2175,29 +2298,42 @@ class Map(DOMWidget, InteractMixin):
 
         if self.zoom_control:
             self.zoom_control_instance = ZoomControl()
-            self.add_control(self.zoom_control_instance)
+            self.add(self.zoom_control_instance)
 
         if self.attribution_control:
             self.attribution_control_instance = AttributionControl(position='bottomright')
-            self.add_control(self.attribution_control_instance)
+            self.add(self.attribution_control_instance)
 
     @observe('zoom_control')
     def observe_zoom_control(self, change):
         if change['new']:
             self.zoom_control_instance = ZoomControl()
-            self.add_control(self.zoom_control_instance)
+            self.add(self.zoom_control_instance)
         else:
             if self.zoom_control_instance is not None and self.zoom_control_instance in self.controls:
-                self.remove_control(self.zoom_control_instance)
+                self.remove(self.zoom_control_instance)
 
     @observe('attribution_control')
     def observe_attribution_control(self, change):
         if change['new']:
             self.attribution_control_instance = AttributionControl(position='bottomright')
-            self.add_control(self.attribution_control_instance)
+            self.add(self.attribution_control_instance)
         else:
             if self.attribution_control_instance is not None and self.attribution_control_instance in self.controls:
-                self.remove_control(self.attribution_control_instance)
+                self.remove(self.attribution_control_instance)
+
+    @validate('panes')
+    def _validate_panes(self, proposal):
+        '''Validate panes.
+        '''
+        error_msg = "Panes should look like: {'pane_name': {'zIndex': 650, 'pointerEvents': 'none'}, ...}"
+        for k1, v1 in proposal.value.items():
+            if not isinstance(k1, str) or not isinstance(v1, dict):
+                raise PaneException(error_msg)
+            for k2, v2 in v1.items():
+                if not isinstance(k2, str) or not isinstance(v2, (str, int, float)):
+                    raise PaneException(error_msg)
+        return proposal.value
 
     _layer_ids = List()
 
@@ -2216,32 +2352,38 @@ class Map(DOMWidget, InteractMixin):
     def add_layer(self, layer):
         """Add a layer on the map.
 
+        .. deprecated :: 0.17.0
+           Use add method instead.
+
         Parameters
         ----------
         layer: Layer instance
             The new layer to add.
         """
-        if isinstance(layer, dict):
-            layer = basemap_to_tiles(layer)
-        if layer.model_id in self._layer_ids:
-            raise LayerException('layer already on map: %r' % layer)
-        self.layers = tuple([layer for layer in self.layers] + [layer])
+        warnings.warn("add_layer is deprecated, use add instead", DeprecationWarning)
+        self.add(layer)
 
     def remove_layer(self, rm_layer):
         """Remove a layer from the map.
 
+        .. deprecated :: 0.17.0
+           Use remove method instead.
+
         Parameters
         ----------
         layer: Layer instance
             The layer to remove.
         """
-        if rm_layer.model_id not in self._layer_ids:
-            raise LayerException('layer not on map: %r' % rm_layer)
-        self.layers = tuple([layer for layer in self.layers if layer.model_id != rm_layer.model_id])
+        warnings.warn("remove_layer is deprecated, use remove instead", DeprecationWarning)
+
+        self.remove(rm_layer)
 
     def substitute_layer(self, old, new):
         """Replace a layer with another one on the map.
 
+        .. deprecated :: 0.17.0
+           Use substitute method instead.
+
         Parameters
         ----------
         old: Layer instance
@@ -2249,14 +2391,19 @@ class Map(DOMWidget, InteractMixin):
         new: Layer instance
             The new layer to add.
         """
-        if isinstance(new, dict):
-            new = basemap_to_tiles(new)
-        if old.model_id not in self._layer_ids:
-            raise LayerException('Could not substitute layer: layer not on map.')
-        self.layers = tuple([new if layer.model_id == old.model_id else layer for layer in self.layers])
+        warnings.warn("substitute_layer is deprecated, use substitute instead", DeprecationWarning)
+
+        self.substitute(old, new)
 
     def clear_layers(self):
-        """Remove all layers from the map."""
+        """Remove all layers from the map.
+
+        .. deprecated :: 0.17.0
+           Use add method instead.
+
+        """
+        warnings.warn("clear_layers is deprecated, use clear instead", DeprecationWarning)
+
         self.layers = ()
 
     controls = Tuple().tag(trait=Instance(Control), sync=True, **widget_serialization)
@@ -2277,29 +2424,42 @@ class Map(DOMWidget, InteractMixin):
     def add_control(self, control):
         """Add a control on the map.
 
+        .. deprecated :: 0.17.0
+           Use add method instead.
+
         Parameters
         ----------
         control: Control instance
             The new control to add.
         """
-        if control.model_id in self._control_ids:
-            raise ControlException('control already on map: %r' % control)
-        self.controls = tuple([c for c in self.controls] + [control])
+
+        warnings.warn("add_control is deprecated, use add instead", DeprecationWarning)
+
+        self.add(control)
 
     def remove_control(self, control):
         """Remove a control from the map.
 
+        .. deprecated :: 0.17.0
+           Use remove method instead.
+
         Parameters
         ----------
         control: Control instance
             The control to remove.
         """
-        if control.model_id not in self._control_ids:
-            raise ControlException('control not on map: %r' % control)
-        self.controls = tuple([c for c in self.controls if c.model_id != control.model_id])
+        warnings.warn("remove_control is deprecated, use remove instead", DeprecationWarning)
+
+        self.remove(control)
 
     def clear_controls(self):
-        """Remove all controls from the map."""
+        """Remove all controls from the map.
+
+        .. deprecated :: 0.17.0
+           Use clear method instead.
+        """
+        warnings.warn("clear_controls is deprecated, use clear instead", DeprecationWarning)
+
         self.controls = ()
 
     def save(self, outfile, **kwargs):
@@ -2315,24 +2475,82 @@ class Map(DOMWidget, InteractMixin):
         embed_minimal_html(outfile, views=[self], **kwargs)
 
     def __iadd__(self, item):
-        if isinstance(item, Layer):
-            self.add_layer(item)
-        elif isinstance(item, Control):
-            self.add_control(item)
+        self.add(item)
         return self
 
     def __isub__(self, item):
-        if isinstance(item, Layer):
-            self.remove_layer(item)
-        elif isinstance(item, Control):
-            self.remove_control(item)
+        self.remove(item)
         return self
 
     def __add__(self, item):
+        return self.add(item)
+
+    def add(self, item):
+        """Add an item on the map: either a layer or a control.
+
+        Parameters
+        ----------
+        item: Layer or Control instance
+            The layer or control to add.
+        """
         if isinstance(item, Layer):
-            self.add_layer(item)
+            if isinstance(item, dict):
+                item = basemap_to_tiles(item)
+            if item.model_id in self._layer_ids:
+                raise LayerException('layer already on map: %r' % item)
+            self.layers = tuple([layer for layer in self.layers] + [item])
+
         elif isinstance(item, Control):
-            self.add_control(item)
+            if item.model_id in self._control_ids:
+                raise ControlException('control already on map: %r' % item)
+            self.controls = tuple([control for control in self.controls] + [item])
+        return self
+
+    def remove(self, item):
+        """Remove an item from the map : either a layer or a control.
+
+        Parameters
+        ----------
+        item: Layer or Control instance
+            The layer or control to remove.
+        """
+        if isinstance(item, Layer):
+            if item.model_id not in self._layer_ids:
+                raise LayerException('layer not on map: %r' % item)
+            self.layers = tuple([layer for layer in self.layers if layer.model_id != item.model_id])
+
+        elif isinstance(item, Control):
+            if item.model_id not in self._control_ids:
+                raise ControlException('control not on map: %r' % item)
+            self.controls = tuple([control for control in self.controls if control.model_id != item.model_id])
+        return self
+
+    def clear(self):
+        "Clear all layers and controls."
+        self.controls = ()
+        self.layers = ()
+        return self
+
+    def substitute(self, old, new):
+        """Replace an item (layer or control) with another one on the map.
+
+        Parameters
+        ----------
+        old: Layer or control instance
+            The old layer or control to remove.
+        new: Layer or control instance
+            The new layer or control to add.
+        """
+        if (isinstance(new, Layer)):
+            if isinstance(new, dict):
+                new = basemap_to_tiles(new)
+            if old.model_id not in self._layer_ids:
+                raise LayerException('Could not substitute layer: layer not on map.')
+            self.layers = tuple([new if layer.model_id == old.model_id else layer for layer in self.layers])
+        elif (isinstance(new, Control)):
+            if old.model_id not in self._control_ids:
+                raise ControlException('Could not substitute control: control not on map.')
+            self.controls = tuple([new if control.model_id == old.model_id else control for control in self.controls])
         return self
 
     # Event handling
