@@ -15,10 +15,11 @@ export class LeafletTileLayerModel extends rasterlayer.LeafletRasterLayerModel {
       url: 'https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png',
       min_zoom: 0,
       max_zoom: 18,
+      min_native_zoom: null,
+      max_native_zoom: null,
       bounds: null,
       tile_size: 256,
-      attribution:
-        'Map data (c) <a href="https://openstreetmap.org">OpenStreetMap</a> contributors',
+      attribution: null,
       detect_retina: false,
       no_wrap: false,
       tms: false,
